/* telnet.i - TELNET Protocal

    Description:    Using the telnet protocol, connects to a Unix system under 
                    program control, executes commands on the Unix system and
                    can optional capture the output to a file.

                    This code should be considered as still in development, 
                    as such it has been unit tested only. Tested with 
                    Progress 9.1B. Requires sockets.
                    
                    Use this software at your own risk!
                    
                    For the full TELNET specification look for RFC854 on 
                    http://www.faqs.org
                    
    Author:         Roger Sabin 05/27/2002
    
    Use Example:    {telnet.i}
    
                    run telnet-exec ("computer IP or name",
                                     "login",
                                     "password",
                                     "$ ",      /* Unix prompt */
                                     "ps -aef",
                                     "tmpfile.txt").

                    Return value = "" to indicate no error, otherwise will give
                    error message.
                    
    Notes:          One of the problems I had with automating the execution
                    of commands on the Unix system was determining when the
                    command line was prompting for another command. The route
                    I chose was to check the beginning of a line for a "prompt"
                    string. Because of this, you should make the prompt string 
                    as unique as possible for the Unix system. For example, 
                    some Unix systems use the computer name at the beginning of 
                    the prompt. This makes for a good prompt string. Others use
                    just '$ ' (note the trailing blank). I have tested this 
                    and it works (as long as no command generates a line that
                    begins with the same string as the prompt).
*/

define variable telnet-socket as handle no-undo.   
define variable telnet-login as char no-undo.
define variable telnet-password as char no-undo.
define variable telnet-stop-flag as log no-undo.
define variable telnet-iac-flag as int no-undo init 0.      
define variable telnet-255-flag as log no-undo init false.
define variable telnet-response as char no-undo init "".
define variable telnet-prompt as char no-undo init "".
define variable telnet-seq-flag as int no-undo init 0.  /* Seq = 0 (connect), 1 (login), 2 (password), 3 (command), 4 (exit), 5 (disconnect) */
define variable telnet-command as char no-undo init "".
define variable telnet-command-crlf as char no-undo init "".
define variable telnet-output-flag as log no-undo init false.

define stream telnet-stm.

procedure telnet-exec:
  define input parameter prm-computer-id as char no-undo.
  define input parameter prm-login as char no-undo.
  define input parameter prm-password as char no-undo.
  define input parameter prm-prompt as char no-undo.    /* beginning part of prompt or entire prompt */
  define input parameter prm-command as char no-undo.
  define input parameter prm-file-name as char no-undo. /* optional - leave blank if output not needed */
  
  define variable v-log as log no-undo.
  define variable v-char-ptr as memptr no-undo.
    
  assign telnet-login        = prm-login
         telnet-password     = prm-password
         telnet-prompt       = prm-prompt
         telnet-command      = prm-command
         telnet-command-crlf = replace(prm-command, chr(10), chr(13) + chr(10))
         telnet-output-flag  = (prm-file-name <> "")
         .
         
  create socket telnet-socket.
  telnet-socket:set-read-response-procedure("telnet-read-handler", this-procedure).
  v-log = telnet-socket:connect("-H " + prm-computer-id + " -S 23") no-error.
  if not v-log then 
    do:
    delete object telnet-socket.
    return "Error - Unable to connect".
    end.
        
  if telnet-output-flag then output stream telnet-stm to value(prm-file-name).
  assign telnet-seq-flag  = 0
         telnet-stop-flag = false
         .
  do while not telnet-stop-flag and telnet-seq-flag < 5
           on error   undo, leave
           on end-key undo, leave:
    assign telnet-stop-flag = true
           telnet-response  = ""
           .
    wait-for read-response of telnet-socket pause 60.   /* for just in case, seems to work */
    if telnet-response <> "" then
        assign set-size(v-char-ptr)      = length(telnet-response) + 1
               put-string(v-char-ptr, 1) = telnet-response
               v-log                     = telnet-socket:write(v-char-ptr, 1, length(telnet-response))
               set-size(v-char-ptr)      = 0
               .
    process events.
  end.
  if telnet-output-flag then 
    do:
    put stream telnet-stm control chr(13) + chr(10).
    output stream telnet-stm close.
    end.
  telnet-socket:disconnect().
  delete object telnet-socket.
  
  return "".
end.



procedure telnet-read-handler:      /* internal use only */
  define variable v-memptr as memptr no-undo.
  define variable v-i as int no-undo.
  define variable v-j as int no-undo.
  define variable v-k as int no-undo.
  define variable v-log as log no-undo.
  define variable v-char as char no-undo.
  define variable v-char1 as char no-undo.
    
  if not telnet-socket:connected() then return.
  telnet-stop-flag = false.
  
  v-k = telnet-socket:get-bytes-available().
  if v-k = 0 then return.
  
  assign set-size(v-memptr) = v-k + 1
         v-log              = telnet-socket:read(v-memptr, 1, v-k, 1)
         v-char             = get-string(v-memptr, 1)
         set-size(v-memptr) = 0
         .
  do v-i = 1 to length(v-char):
    assign v-char1 = substring(v-char, v-i, 1)
           v-j     = asc(v-char1)
           .
    if v-j = 13 and telnet-iac-flag = 0 then next.
    if v-j = 255 then
        do:
        if telnet-255-flag then /* double 255 chars -> keep just the first one */
            do:
            telnet-255-flag = false.
            next.
            end.
        telnet-255-flag = true.
        if telnet-iac-flag = 0 then
            assign telnet-iac-flag = 1
                   telnet-response = telnet-response + v-char1
                   .
        end.
    else 
        do:
        telnet-255-flag = false.
        if telnet-iac-flag = 1 then
            do:
            telnet-iac-flag = 2.
            if v-j = 253 then 
                telnet-response = telnet-response + chr(252). /* do -> won't */
            else
                assign telnet-iac-flag = 0
                       telnet-response = substring(telnet-response, 1, length(telnet-response) - 1)
                       .
            end.
        else if telnet-iac-flag = 2 then
            assign telnet-iac-flag = 0
                   telnet-response = telnet-response + v-char1
                   .
        end.
  end.
  if telnet-seq-flag = 0 and index(v-char, "login:") > 0 then 
    assign telnet-response = telnet-response + telnet-login + chr(10)
           telnet-seq-flag = 1
           .
  else if telnet-seq-flag = 1 and index(v-char, "Password:") > 0 then 
    assign telnet-response = telnet-response + telnet-password + chr(10)
           telnet-seq-flag = 2
           .
  else if telnet-seq-flag = 2 and (v-char begins telnet-prompt or index(v-char, chr(10) + telnet-prompt) > 0) then
    assign telnet-response = telnet-response + telnet-command + chr(10)
           telnet-seq-flag = 3
           .
  else if telnet-seq-flag = 3 or telnet-seq-flag = 4 then
    do:
    if v-char begins telnet-command-crlf then v-char = substring(v-char, length(telnet-command-crlf) + 1).
    v-i = r-index(v-char, chr(10) + telnet-prompt).
    if v-i > 0 or (v-i = 0 and v-char <> "" and v-char begins telnet-prompt) then 
        do:
        assign v-char  = substring(v-char, 1, v-i)
               v-char1 = trim(v-char, chr(10) + chr(13))
               .
        if telnet-seq-flag = 3 then 
            telnet-seq-flag = 4.
        else
            assign telnet-response  = telnet-response + "exit" + chr(10)
                   telnet-seq-flag  = 5
                   .
        end.
    else
        do:
        v-char1 = trim(v-char, chr(10) + chr(13)).
        if v-char1 <> "" then telnet-seq-flag = 4.
        end.
    if telnet-output-flag and v-char1 <> "" then put stream telnet-stm control replace(v-char, chr(11), "").
    end.
  process events.   /* do not remove - needed to make sure we process exit command (don't know why it works, but it does) */
end.
